from typing import Dict, Optional, List
from functools import wraps
import json
import os
from enum import Enum
import requests
import datetime
import sys
from pydantic import BaseModel
from typing import Any, Union, Literal
from playwright.sync_api import sync_playwright, Browser, BrowserContext, Page, ElementHandle
from playwright.async_api import async_playwright, Browser as AsyncBrowser, BrowserContext as AsyncBrowserContext, Page as AsyncPage, ElementHandle as AsyncElementHandle
from dotenv import load_dotenv

class FinicEnvironment(str, Enum):
    LOCAL = "local"
    PROD = "production"
    DEV = "development"


class LogSeverity(str, Enum):
    DEFAULT = "DEFAULT"
    WARNING = "WARNING"
    ERROR = "ERROR"


class ExecutionLog(BaseModel):
    severity: LogSeverity
    message: str
    timestamp: Optional[datetime.datetime] = None

    class Config:
        json_encoders = {datetime: lambda v: v.isoformat() if v else None}


class ExecutionAttempt(BaseModel):
    success: bool
    attempt_number: int
    logs: List[ExecutionLog] = []


class LogExecutionAttemptRequest(BaseModel):
    execution_id: str
    agent_id: str
    results: Dict[str, Any]
    attempt: ExecutionAttempt


class StdoutLogger:
    def __init__(self, original_stdout):
        self.logs: List[ExecutionLog] = []
        self.original_stdout = original_stdout

    def write(self, message):
        if message.strip():  # Avoid logging empty messages
            log = ExecutionLog(
                severity=LogSeverity.DEFAULT,
                timestamp=datetime.datetime.now(datetime.timezone.utc),
                message=message.strip(),
            )
            self.logs.append(log)
            self.original_stdout.write(
                f"{log.timestamp} [{log.severity.value}] {log.message}\n"
            )

    def flush(self):
        pass

    def get_logs(self):
        return self.logs

def on_browser_disconnected(browser_context: Union[BrowserContext, AsyncBrowserContext]):
    print("Browser disconnected")
    storage_state_path = os.path.join(os.getcwd(), f"storage_state.json")
    browser_context.storage_state(path=storage_state_path)
    print(f"Browser context saved to: {storage_state_path}")

class Finic:
    api_key: Optional[str] = None
    env: Optional[FinicEnvironment] = None
    context_storage_path: Optional[str] = "storage_state.json"

    def __init__(
        self,
        api_key: Optional[str] = None,
        environment: Optional[FinicEnvironment] = None,
        url: Optional[str] = None
    ):

        default_env = os.environ.get("FINIC_ENVIRONMENT") or FinicEnvironment.LOCAL
        self.environment = environment if environment else FinicEnvironment(default_env)
        if api_key:
            self.api_key = api_key
        else:
            self.api_key = os.getenv("FINIC_API_KEY")
        self.url = url or "https://api.finic.io"

    def get_agent_input(self) -> Optional[Dict[str, Any]]:
        load_dotenv()
        if os.getenv("FINIC_INPUT"):
            return json.loads(os.getenv("FINIC_INPUT"))
        return None
    
    def save_browser_context(self, context: BrowserContext, browser_id: Optional[str] = None):
        if self.environment == FinicEnvironment.LOCAL:
            path = os.path.join(os.getcwd(), f"browser_state.json")
            context.storage_state(path=path)
            print(f"Browser context saved to: {path}")
        else:
            
            browser_id = browser_id or os.getenv("FINIC_BROWSER_ID")
            if not browser_id:
                return
            browser_context = context.storage_state()
            response = requests.post(
                f"{self.url}/browser-state/{browser_id}",
                headers={"Authorization": f"Bearer {self.api_key}"},
                json=browser_context
            )
            response.raise_for_status()
            browser = response.json()
            print(f"Browser context saved for browser: {browser['id']}")

    def get_browser_context(self, browser_id: Optional[str] = None) -> Optional[Dict[str, Any]]:
        if self.environment == FinicEnvironment.LOCAL:
            path = os.path.join(os.getcwd(), f"browser_state.json")
            if os.path.exists(path):
                with open(path, 'r') as f:
                    return json.load(f)
            else:
                return None
        else:
            browser_id = browser_id or os.getenv("FINIC_BROWSER_ID")
            if not browser_id:
                return None
            response = requests.get(
                f"{self.url}/browser-state/{browser_id}",
                headers={"Authorization": f"Bearer {self.api_key}"},
            )
            response.raise_for_status()
            browser = response.json()
            if browser and browser.get('state'):
                return browser.get('state')
            else:
                return None
        
    def save_session_results(self, results: Dict):
        if self.environment == FinicEnvironment.LOCAL:
            path = os.path.join(os.getcwd(), f"results.json")
            with open(path, 'w') as f:
                json.dump(results, f, ensure_ascii=False, indent=4, default=lambda o: o.dict())
            print(f"Session results saved to: {path}")
        else:
            session_id = os.getenv("FINIC_SESSION_ID")
            response = requests.post(
                f"{self.url}/session-results/{session_id}",
                headers={"Authorization": f"Bearer {self.api_key}"},
                json=results
            )
            response.raise_for_status()
            print(f"Session results saved for session: {session_id}")

    def get_session_results(self, session_id: Optional[str] = None) -> Optional[Dict]:
        if self.environment == FinicEnvironment.LOCAL:
            path = os.path.join(os.getcwd(), f"results.json")
            if os.path.exists(path):    
                with open(path, 'r') as f:
                    return json.load(f)
            else:
                return None
        else:
            if not session_id:
                session_id = os.getenv("FINIC_SESSION_ID")
            response = requests.get(
                f"{self.url}/session-results/{session_id}",
                headers={"Authorization": f"Bearer {self.api_key}"},
            )
            response.raise_for_status()
            return response.json()
    
    def launch_browser_sync(self, browser_type: Optional[Literal["chromium", "firefox", "webkit"]] = "chromium", **kwargs) -> BrowserContext:
        playwright = sync_playwright().start()
        browser = playwright.chromium.launch(**kwargs)
        saved_context = self.get_browser_context()
        context = browser.new_context(storage_state=saved_context)
        context.add_init_script("""
            Object.defineProperty(navigator, 'webdriver', {
                get: () => undefined
            });
        """)

        return context
    
    async def launch_browser_async(self, browser_type: Optional[Literal["chromium", "firefox", "webkit"]] = "chromium", **kwargs) -> AsyncBrowserContext:
        playwright = async_playwright().start()
        browser = await getattr(playwright, browser_type).launch(**kwargs)
        context = await browser.new_context()
        context.add_init_script("""
            Object.defineProperty(navigator, 'webdriver', {
                get: () => undefined
            });
        """)
        context.on("disconnected", on_browser_disconnected)
        
        try:
            if os.path.exists(self.context_storage_path):
                with open(self.context_storage_path, 'r') as f:
                    storage_state = json.load(f)
                    cookies = storage_state.get('cookies', [])
                    context.add_cookies(cookies)
        except Exception as e:
            print(f"Error loading storage state: {e}")

        return context
    
    @classmethod
    def screenshot(self, page: Union[Page, AsyncPage], selector: str, path: str) -> None:
        # Create a new div for screenshot
        page.evaluate("""() => {
            const outerContainer = document.createElement('div');
            outerContainer.id = 'screenshotOuterContainer';
            outerContainer.style.position = 'fixed';
            outerContainer.style.zIndex = '2147483647'; // Maximum z-index value
            outerContainer.style.top = '0';
            outerContainer.style.left = '0';
            outerContainer.style.width = '100%';
            outerContainer.style.height = '100%';
            outerContainer.style.backgroundColor = 'white';
            outerContainer.style.display = 'flex';
            outerContainer.style.justifyContent = 'center';
            outerContainer.style.alignItems = 'center';

            const innerContainer = document.createElement('div');
            innerContainer.id = 'screenshotInnerContainer';
            innerContainer.style.display = 'inline-block';

            outerContainer.appendChild(innerContainer);
            document.body.appendChild(outerContainer);
        }""")

        # Move the ad into the screenshot container
        cloned_element = page.evaluate_handle("""(selector) => {
            const innerContainer = document.getElementById('screenshotInnerContainer');
            const adElement = document.evaluate(selector, document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
            
            if (innerContainer && adElement) {
                const clone = adElement.cloneNode(true);
                const styles = window.getComputedStyle(adElement);
                
                clone.style.cssText = Array.from(styles).reduce((str, property) => {
                    return `${str}${property}:${styles.getPropertyValue(property)};`;
                }, '');
                
                // Reset positioning styles
                clone.style.position = 'static';
                clone.style.margin = '0';
                clone.style.width = 'auto';
                clone.style.height = 'auto';
                
                innerContainer.appendChild(clone);
                return clone;
            }
            return null;
        }""", selector)

        page.wait_for_selector(selector)  # Wait for the element to be added
        cloned_element.screenshot(path=path)

        # Remove the screenshot container div
        page.evaluate("""() => {
            const outerContainer = document.getElementById('screenshotOuterContainer');
            outerContainer.remove();
        }""")

    @classmethod
    def entrypoint(cls, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            config = {}
            if os.path.exists('finic_config.json'):
                with open('finic_config.json', 'r') as config_file:
                    config = json.load(config_file)
            inputs = config.get("args", {})

            result = func(*args, **kwargs, **inputs)

            with open('output.json', 'w') as output_file:
                json.dump(result, output_file)

            return result
        return wrapper

    @classmethod
    def procedure(cls, func):
        @wraps(func)
        def wrapper(page, *args, **kwargs):
            if not isinstance(page, Page):
                raise TypeError("The first argument must be a Page object")
            result = func(page, *args, **kwargs)
            if not isinstance(result, Page):
                raise TypeError("The function must return a Page object")
            return result
        return wrapper

    @classmethod
    def max_scroll(self, element: Union[Page, ElementHandle], timeout: int = 2000):
        last_height = element.evaluate('document.body.scrollHeight')
        while True:
            element.evaluate('window.scrollTo(0, document.body.scrollHeight)')
            element.wait_for_timeout(timeout)
            new_height = element.evaluate('document.body.scrollHeight')
            if new_height == last_height:
                break
            last_height = new_height
        return new_height
    
    def deploy_agent(self, agent_id: str, agent_name: str, num_retries: int, zip_file: str) -> bool:
        with open(zip_file, 'rb') as f:
            upload_file = f.read()
        response = requests.post(
            f"{self.url}/agent-upload-link/{agent_id}",
            headers={"Authorization": f"Bearer {self.api_key}"},
            json={"agent_id": agent_id, "agent_name": agent_name, "num_retries": num_retries}
        )
        response.raise_for_status()
        upload_url = response.json()["upload_url"]
        response = requests.put(upload_url, data=upload_file)
        response.raise_for_status()
        return True
    
    



from typing import Optional
from playwright.sync_api import Response


def get_cookies_from_string(cookies_string: str, url: Optional[str] = None, domain: Optional[str] = None, path: Optional[str] = None):
  # Split the content by ";"
  pairs = cookies_string.split("; ")

  cookies = []

  # Populate the dictionary
  for pair in pairs:
      if "=" in pair:
          key, value = pair.split("=", 1)
          cookie = {"name": key, "value": value}
          if url:
              cookie["domain"] = url
          elif domain and path:
              cookie["domain"] = domain
              cookie["path"] = path
          else:
              raise ValueError("Either url, or domain and path must be provided")
          cookies.append(cookie)

  return cookies

def handle_response(response: Response):
    pass

def generate_selector(element: NodeInfo):
    pass


###############################################################################
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
###############################################################################

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code.
#
# ruff: noqa: E501,F401
# flake8: noqa: E501,F401
# pylint: disable=unused-import,line-too-long
# fmt: off
import baml_py
from enum import Enum
from pydantic import BaseModel, ConfigDict
from typing import Dict, List, Optional, Union


class SelectorStyle(str, Enum):
    
    ID_BASED = "ID_BASED"
    CLASS_BASED = "CLASS_BASED"
    PATH_BASED = "PATH_BASED"
    INNER_TEXT_BASED = "INNER_TEXT_BASED"
    LABEL_BASED = "LABEL_BASED"
    SIBLING_BASED = "SIBLING_BASED"
    HREF_BASED = "HREF_BASED"
    STYLE_BASED = "STYLE_BASED"
    OTHER_ATTRIBUTE_BASED = "OTHER_ATTRIBUTE_BASED"

class Attribute(BaseModel):
    
    
    name: str
    value: str

class Element(BaseModel):
    
    
    tagName: str
    textContent: Optional[str] = None
    attributes: List["Attribute"]

class PlanningStep(BaseModel):
    
    
    step_number: int
    description: str

class Step(BaseModel):
    
    
    step_index: int
    selector: str
    instructions: str



###############################################################################
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
###############################################################################

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code.
#
# ruff: noqa: E501,F401
# flake8: noqa: E501,F401
# pylint: disable=unused-import,line-too-long
# fmt: off
from . import types
from . import tracing
from . import partial_types
from .globals import reset_baml_env_vars


from .sync_client import b



__all__ = [
  "b",
  "partial_types",
  "tracing",
  "types",
  "reset_baml_env_vars",
]


###############################################################################
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
###############################################################################

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code.
#
# ruff: noqa: E501,F401
# flake8: noqa: E501,F401
# pylint: disable=unused-import,line-too-long
# fmt: off
import baml_py
from enum import Enum
from pydantic import BaseModel, ConfigDict
from typing import Dict, List, Optional, Union

from . import types

###############################################################################
#
#  These types are used for streaming, for when an instance of a type
#  is still being built up and any of its fields is not yet fully available.
#
###############################################################################


class Attribute(BaseModel):
    
    
    name: Optional[str] = None
    value: Optional[str] = None

class Element(BaseModel):
    
    
    tagName: Optional[str] = None
    textContent: Optional[str] = None
    attributes: List["Attribute"]

class PlanningStep(BaseModel):
    
    
    step_number: Optional[int] = None
    description: Optional[str] = None

class Step(BaseModel):
    
    
    step_index: Optional[int] = None
    selector: Optional[str] = None
    instructions: Optional[str] = None



from playwright.async_api import async_playwright, Playwright, Page, ElementHandle, CDPSession
import asyncio
from enum import Enum
from typing import List, Dict, Any, Optional
from bs4 import BeautifulSoup, Tag
import shutil
from pydantic import BaseModel
from .baml_client import b as baml
from .baml_client.types import Element, Step
from baml_py import ClientRegistry
import uuid
import json
from typing import Literal

client_registry = ClientRegistry()

class NodeInfo(BaseModel):
    tagName: str
    backend_id: int
    user_assigned_id: Optional[str] = str(uuid.uuid4())
    className: str
    textContent: str
    outerHTML: str

def print_welcome_message():
    # Clear the terminal
    print("\033[2J\033[H", end="")
    print("\033[38;2;255;81;6m" + """
    ███████╗  ██╗  ███╗   ██╗  ██╗  ██████╗
    ██╔════╝  ██║  ████╗  ██║  ██║  ██╔════╝
    █████╗    ██║  ██╔██╗ ██║  ██║  ██║     
    ██╔══╝    ██║  ██║╚██╗██║  ██║  ██║     
    ██║       ██║  ██║ ╚████║  ██║  ╚██████╗
    ╚═╝       ╚═╝  ╚═╝  ╚═══╝  ╚═╝   ╚═════╝
    """ + "\033[0m")
    print("\033[1m\033[38;2;255;165;0mGenerate code to automate any website using AI.\033[0m\n")
    print("Type 'help' for a list of commands. View full instructions here: https://docs.finic.ai/capture-mode")

def print_command_prompt():
    print("\n\033[1m\033[38;2;255;165;0mWhat should I do with this element? For example, \"click to open in a new tab, then copy the url in the new tab, then close the tab\"\033[0m")
    print("\033[1m\033[38;2;255;165;0mPrompt:\033[0m ", end="", flush=True)

async def handle_add_node(instructions: str, step_index: int, current_node: List[NodeInfo], cdp_session: CDPSession, page: Page) -> Step:
    target_node = await cdp_session.send("DOM.resolveNode", {"backendNodeId": current_node[0].backend_id})
    target_object_id = target_node["object"]["objectId"]

    target_element = await cdp_session.send("Runtime.callFunctionOn", {
        "functionDeclaration": """
        function() {
            return {
                tagName: this.tagName,
                id: this.id,
                className: this.className,
                attributes: Array.from(this.attributes).map(attr => ({name: attr.name, value: attr.value})),
                textContent: this.textContent
            };
        }
        """,
        "objectId": target_object_id,
        "returnByValue": True
    })

    target = Element(
        tagName=target_element["result"]["value"]["tagName"],
        textContent=target_element["result"]["value"]["textContent"],
        attributes=target_element["result"]["value"]["attributes"]
    )
    
    next_sibling_element = await cdp_session.send("Runtime.callFunctionOn", {
        "functionDeclaration": """
        function() {
            if (this.nextElementSibling) {
                return {
                    tagName: this.nextElementSibling.tagName,
                    id: this.nextElementSibling.id,
                    className: this.nextElementSibling.className,
                    attributes: Array.from(this.nextElementSibling.attributes).map(attr => ({name: attr.name, value: attr.value})),
                    textContent: this.nextElementSibling.textContent
                };
            } else {
                return null;
            }
        }
        """,
        "objectId": target_object_id,
        "returnByValue": True
    })

    previous_sibling_element = await cdp_session.send("Runtime.callFunctionOn", {
        "functionDeclaration": """
        function() {
            if (this.previousElementSibling) {
                return {
                    tagName: this.previousElementSibling.tagName,
                    id: this.previousElementSibling.id,
                    className: this.previousElementSibling.className,
                    attributes: Array.from(this.previousElementSibling.attributes).map(attr => ({name: attr.name, value: attr.value})),
                    textContent: this.previousElementSibling.textContent
                };
            } else {
                return null;
            }
        }
        """,
        "objectId": target_object_id,
        "returnByValue": True
    })
    
    ancestors = await cdp_session.send("Runtime.callFunctionOn", {
        "functionDeclaration": """
        function() {
            let ancestors = [];
            let currentNode = this;
            while (currentNode && currentNode.tagName !== 'BODY') {
                currentNode = currentNode.parentNode;
                if (currentNode) {
                    ancestors.push({
                        tagName: currentNode.tagName,
                        id: currentNode.id,
                        className: currentNode.className,
                        innerHTML: currentNode.innerHTML,
                        attributes: Array.from(currentNode.attributes).map(attr => ({name: attr.name, value: attr.value}))
                    });
                }
            }
            return ancestors;
        }
        """,
        "objectId": target_object_id,
        "returnByValue": True
    })
    next_sibling = None
    if next_sibling_element["result"]["value"] is not None:
        next_sibling = Element(
            tagName=next_sibling_element["result"]["value"]["tagName"],
            textContent=next_sibling_element["result"]["value"]["textContent"],
            attributes=next_sibling_element["result"]["value"]["attributes"]
        )
    previous_sibling = None
    if previous_sibling_element["result"]["value"] is not None:
        previous_sibling = Element(
            tagName=previous_sibling_element["result"]["value"]["tagName"],
            textContent=previous_sibling_element["result"]["value"]["textContent"],
            attributes=previous_sibling_element["result"]["value"]["attributes"]
        )

    trimmed_ancestors = ""
    for i, ancestor in enumerate(ancestors["result"]["value"][:10]):
        if ancestor["tagName"] == "BODY":
            continue
        indent = "  " * i
        attributes = " ".join([f'{attr["name"]}="{attr["value"]}"' for attr in ancestor["attributes"]])
        trimmed_ancestors += f"{indent}<{ancestor['tagName'].lower()} {attributes}>\n"
    trimmed_ancestors = trimmed_ancestors.rstrip()
    import pdb; pdb.set_trace()
    selector_styles = baml.DetermineSelectorStyle(instructions=instructions, target=target, next_sibling=next_sibling, previous_sibling=previous_sibling, ancestors=trimmed_ancestors, baml_options={"client_registry": client_registry})
    if "ID_BASED" in selector_styles and "id" not in target.attributes:
        selector_styles.remove("ID_BASED")
    if "CLASS_BASED" in selector_styles and "class" not in target.attributes:
        selector_styles.remove("CLASS_BASED")
    if "INNER_TEXT_BASED" in selector_styles and (target.textContent == "" or target.textContent is None):
        selector_styles.remove("INNER_TEXT_BASED")
    if "SIBLING_BASED" in selector_styles and next_sibling is None and previous_sibling is None:
        selector_styles.remove("SIBLING_BASED")
    selectors = baml.GenerateXPathSelector(
        instructions=instructions, 
        target=target, 
        next_sibling=next_sibling, 
        previous_sibling=previous_sibling, 
        ancestors=trimmed_ancestors if "PATH_BASED" in selector_styles else None, 
        selector_styles=selector_styles,
        baml_options={"client_registry": client_registry}
    )

    chosen_selector = None
    # Test the selectors
    for selector in selectors:
        element = await page.query_selector(selector)
        if element is not None:
            chosen_selector = selector
            break

    if chosen_selector is None:
        print("No generated selectors were valid. Please enter a selector manually:")
        chosen_selector = input()
    print(f"Element processed. Choose another element or type 'g|generate' to generate selectors for queued elements:")
    
    return Step(step_index=step_index, selector=chosen_selector, instructions=instructions)
    

def print_element_to_terminal(result: NodeInfo):
    terminal_width, _ = shutil.get_terminal_size()
    box_width = min(80, terminal_width - 2)  # Max 80, or 2 less than terminal width

    # Print the box
    print("\033[1m\033[94m┌─" + "─" * (box_width - 2) + "┐\033[0m")
    print(f"\033[1m\033[94m│\033[0m {'Currently selected element':^{box_width-2}}\033[1m\033[94m│\033[0m")
    print("\033[1m\033[94m├─" + "─" * (box_width - 2) + "┤\033[0m")
    print(f"\033[1m\033[94m│\033[0m \033[1m\033[92mTag:\033[0m {result.tagName:<{box_width-7}}\033[1m\033[94m│\033[0m")
    print(f"\033[1m\033[94m│\033[0m \033[1m\033[92mID:\033[0m {result.backend_id:<{box_width-6}}\033[1m\033[94m│\033[0m")
    print(f"\033[1m\033[94m│\033[0m \033[1m\033[92mClass:\033[0m {result.className:<{box_width-9}}\033[1m\033[94m│\033[0m")
    print(f"\033[1m\033[94m│\033[0m \033[1m\033[92mText content:\033[0m{' ':{box_width-15}}\033[1m\033[94m│\033[0m")
    text_content = result.textContent[:box_width-4]  # Truncate to fit
    print(f"\033[1m\033[94m│\033[0m {text_content:<{box_width-2}}\033[1m\033[94m│\033[0m")
    print("\033[1m\033[94m└─" + "─" * (box_width - 2) + "┘\033[0m")

async def generate_procedure(
        procedure_name: str,
        procedure_steps: List[Step],
        llm_provider: Literal["openai", "anthropic"], 
        provider_api_key: str, 
        cdp_session: CDPSession, 
        node_queue: List[NodeInfo]):
    
    procedure_code = await baml.GeneratePlaywrightCode(procedure_name, procedure_steps)
    procedure_filename = f"procedures/{procedure_name}.py"
    with open(procedure_filename, "w") as f:
        f.write(procedure_code)
    
    print(f"\nProcedure code has been written to {procedure_filename}")

async def handle_inspect_node(cdp_session: CDPSession, page: Page, event: Dict[str, Any], current_node: List[NodeInfo]):
    backend_node_id = event["backendNodeId"]

    # Request the document so DOM.pushNodesByBackendIdsToFrontend works
    document = await cdp_session.send("DOM.getDocument")
    
    # Find the node in the DOM
    await cdp_session.send("DOM.pushNodesByBackendIdsToFrontend", {"backendNodeIds": [backend_node_id]})
    node_details = await cdp_session.send("DOM.describeNode", {"backendNodeId": backend_node_id})
    node_id = node_details["node"]["nodeId"]
    await cdp_session.send("DOM.setInspectedNode", {"nodeId": node_id})

    # Navigate to the node in the devtools elements inspector
    await cdp_session.send('DOM.setInspectedNode', {
        'nodeId': node_id
    })
    outer_html = await cdp_session.send("DOM.getOuterHTML", {"nodeId": node_id})
    result = NodeInfo(
        tagName=node_details["node"]["localName"],
        backend_id=backend_node_id,
        className=" ".join([value for attr, value in zip(node_details["node"].get("attributes", [])[::2], node_details["node"].get("attributes", [])[1::2]) if attr == "class"]),
        textContent=BeautifulSoup(outer_html["outerHTML"], 'html.parser').get_text(strip=True),
        outerHTML=outer_html["outerHTML"]
    )

    current_node[0] = result
    print_welcome_message()
    print_element_to_terminal(result)
    print_command_prompt()

async def capture(llm_provider: Literal["openai", "anthropic"], provider_api_key: str, url: str):
    playwright = await async_playwright().start()
    
    browser = await playwright.chromium.launch(headless=False)
    page = await browser.new_page()

    # Enable CDP Session
    cdp_session = await page.context.new_cdp_session(page)
    await cdp_session.send('DOM.enable')
    await cdp_session.send('Overlay.enable')


    # Enable inspect mode
    interaction_mode = True
    await cdp_session.send('Overlay.setInspectMode', {
        'mode': 'searchForNode',
        'highlightConfig': {'showInfo': True, 'showExtensionLines': True, 'contentColor': {'r': 255, 'g': 81, 'b': 6, 'a': 0.2}}
    })
    
    # Navigate to a website
    await page.goto(url)

    print_welcome_message()
    print("A Chromium browser with devtools enabled should now be open.")

    # Track selected nodes
    procedure_steps: List[Step] = []
    current_node: List[NodeInfo] = [None]

    # Set up event listener for DOM.inspectNodeRequested
    cdp_session.on("Overlay.inspectNodeRequested", lambda event: handle_inspect_node(cdp_session, page, event, current_node))

    # Set up BAML client registry
    client_registry.add_llm_client(name="Sonnet", 
                                   provider=llm_provider, 
                                   options={
                                       "api_key": provider_api_key,
                                       "model": "claude-3-5-sonnet-20240620"
                                    })
    client_registry.set_primary("Sonnet")

    while True:
        user_input = await asyncio.get_event_loop().run_in_executor(None, input)
        print("\033[1A", end="")  # Move cursor up one line
        print("\033[K", end="")   # Clear the line
        if user_input.lower() in ['quit', 'q']:
            break
        elif user_input.lower() in ['generate', 'g']:
            print("Generating procedure...")
            print("Enter a unique name for this procedure:")
            procedure_name = input()
            await generate_procedure(procedure_name, procedure_steps,llm_provider, provider_api_key, cdp_session, node_queue)
            print("Procedure generated.")
            procedure_steps = []
        elif user_input.lower() in ['mode', 'm']:
            if interaction_mode:
                await cdp_session.send('Overlay.setInspectMode', {'mode': 'none', 'highlightConfig': {}})
                interaction_mode = False
                print("Interaction mode disabled.")
            else:
                await cdp_session.send('Overlay.setInspectMode', {'mode': 'searchForNode', 'highlightConfig': {'showInfo': True, 'showExtensionLines': True, 'contentColor': {'r': 255, 'g': 81, 'b': 6, 'a': 0.2}}})
                interaction_mode = True
                print("Selection mode enabled.")
        elif user_input.lower() in ['help', 'h']:
            print("Click on eleements in the browser. Confirm your selection in the box above. Use 'a|add' to queue an element for generation. Use 'g|generate' to generate selectors for queued elements. Use 'l|list' to view queued elements.")
            print("\n\033[1mCommands:\033[0m")
            print("  • \033[1m'add'|'a'\033[0m     - Queue this element for selector generation")
            print("  • \033[1m'list'|'l'\033[0m    - View currently queued elements")
            print("  • \033[1m'm'|'mode'\033[0m    - Change between interaction and selection mode")
            print("  • \033[1m'generate'|'g'\033[0m - Start generating selectors")
            print("  • \033[1m'quit'|'q'\033[0m    - Quit the program")
            print("\n\033[1m\033[38;2;255;165;0mEnter command:\033[0m ", end="", flush=True)
        else:
            result = await handle_add_node(user_input, len(procedure_steps), current_node, cdp_session, page)
            procedure_steps.append(result)
    await browser.close()



###############################################################################
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
###############################################################################

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code.
#
# ruff: noqa: E501,F401
# flake8: noqa: E501,F401
# pylint: disable=unused-import,line-too-long
# fmt: off
import typing
from baml_py.type_builder import FieldType, TypeBuilder as _TypeBuilder, ClassPropertyBuilder, EnumValueBuilder, EnumBuilder, ClassBuilder

class TypeBuilder(_TypeBuilder):
    def __init__(self):
        super().__init__(classes=set(
          ["Attribute","Element","PlanningStep","Step",]
        ), enums=set(
          ["SelectorStyle",]
        ))









__all__ = ["TypeBuilder"]


###############################################################################
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
###############################################################################

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code.
#
# ruff: noqa: E501,F401
# flake8: noqa: E501,F401
# pylint: disable=unused-import,line-too-long
# fmt: off
from .globals import DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_CTX

trace = DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_CTX.trace_fn
set_tags = DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_CTX.upsert_tags
def flush():
  DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_CTX.flush()
on_log_event = DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_CTX.on_log_event


__all__ = ['trace', 'set_tags', "flush", "on_log_event"]